"""NumpyArrayItem.

This module defines the NumpyArrayItem class, which represents a NumPy array item.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy

from skore.item.item import Item


class NumpyArrayItem(Item):
    """
    A class to represent a NumPy array item.

    This class encapsulates a NumPy array along with its creation and update timestamps.

    Attributes
    ----------
    array_list : list
        The list representation of the NumPy array.
    created_at : str
        The timestamp when the item was created, in ISO format.
    updated_at : str
        The timestamp when the item was last updated, in ISO format.

    Methods
    -------
    array() : numpy.ndarray
        Returns the NumPy array representation of the stored list.
    factory(array: numpy.ndarray) : NumpyArrayItem
        Creates a new NumpyArrayItem instance from a NumPy array.
    """

    def __init__(
        self,
        array_list: list,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a NumpyArrayItem.

        Parameters
        ----------
        array_list : list
            The list representation of the NumPy array.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.array_list = array_list

    @cached_property
    def array(self) -> numpy.ndarray:
        """
        Convert the stored list to a NumPy array.

        Returns
        -------
        numpy.ndarray
            The NumPy array representation of the stored list.
        """
        import numpy

        return numpy.asarray(self.array_list)

    @classmethod
    def factory(cls, array: numpy.ndarray) -> NumpyArrayItem:
        """
        Create a new NumpyArrayItem instance from a NumPy array.

        Parameters
        ----------
        array : numpy.ndarray
            The NumPy array to store.

        Returns
        -------
        NumpyArrayItem
            A new NumpyArrayItem instance.
        """
        import numpy

        if not isinstance(array, numpy.ndarray):
            raise TypeError(f"Type '{array.__class__}' is not supported.")

        instance = cls(array_list=array.tolist())

        # add array as cached property
        instance.array = array

        return instance



"""PandasSeriesItem.

This module defines the PandasSeriesItem class,
which represents a pandas Series item.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas

from skore.item.item import Item


class PandasSeriesItem(Item):
    """
    A class to represent a pandas Series item.

    This class encapsulates a pandas Series along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        series_list: list,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a PandasSeriesItem.

        Parameters
        ----------
        series_list : list
            The list representation of the series.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.series_list = series_list

    @cached_property
    def series(self) -> pandas.Series:
        """The pandas Series."""
        import pandas

        return pandas.Series(self.series_list)

    @classmethod
    def factory(cls, series: pandas.Series) -> PandasSeriesItem:
        """
        Create a new PandasSeriesItem instance from a pandas Series.

        Parameters
        ----------
        series : pd.Series
            The pandas Series to store.

        Returns
        -------
        PandasSeriesItem
            A new PandasSeriesItem instance.
        """
        import pandas

        if not isinstance(series, pandas.Series):
            raise TypeError(f"Type '{series.__class__}' is not supported.")

        instance = cls(series_list=series.to_list())

        # add series as cached property
        instance.series = series

        return instance



"""SklearnBaseEstimatorItem.

This module defines the SklearnBaseEstimatorItem class,
which represents a scikit-learn BaseEstimator item.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sklearn.base

from skore.item.item import Item


class SklearnBaseEstimatorItem(Item):
    """
    A class to represent a scikit-learn BaseEstimator item.

    This class encapsulates a scikit-learn BaseEstimator along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        estimator_html_repr: str,
        estimator_skops: bytes,
        estimator_skops_untrusted_types: list[str],
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a SklearnBaseEstimatorItem.

        Parameters
        ----------
        estimator_html_repr : str
            The HTML representation of the scikit-learn estimator.
        estimator_skops : bytes
            The skops representation of the scikit-learn estimator.
        estimator_skops_untrusted_types : list[str]
            The list of untrusted types in the skops representation.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.estimator_html_repr = estimator_html_repr
        self.estimator_skops = estimator_skops
        self.estimator_skops_untrusted_types = estimator_skops_untrusted_types

    @cached_property
    def estimator(self) -> sklearn.base.BaseEstimator:
        """
        Convert the stored skops object to a scikit-learn BaseEstimator.

        Returns
        -------
        sklearn.base.BaseEstimator
            The scikit-learn BaseEstimator representation of the stored skops object.
        """
        import skops.io

        return skops.io.loads(
            self.estimator_skops, trusted=self.estimator_skops_untrusted_types
        )

    @classmethod
    def factory(cls, estimator: sklearn.base.BaseEstimator) -> SklearnBaseEstimatorItem:
        """
        Create a SklearnBaseEstimatorItem instance from a scikit-learn BaseEstimator.

        Parameters
        ----------
        estimator : sklearn.base.BaseEstimator
            The scikit-learn BaseEstimator to store.

        Returns
        -------
        SklearnBaseEstimatorItem
            A new SklearnBaseEstimatorItem instance.
        """
        import sklearn.base
        import sklearn.utils
        import skops.io

        if not isinstance(estimator, sklearn.base.BaseEstimator):
            raise TypeError(f"Type '{estimator.__class__}' is not supported.")

        estimator_html_repr = sklearn.utils.estimator_html_repr(estimator)
        estimator_skops = skops.io.dumps(estimator)
        estimator_skops_untrusted_types = skops.io.get_untrusted_types(
            data=estimator_skops
        )

        instance = cls(
            estimator_html_repr=estimator_html_repr,
            estimator_skops=estimator_skops,
            estimator_skops_untrusted_types=estimator_skops_untrusted_types,
        )

        # add estimator as cached property
        instance.estimator = estimator

        return instance



"""PandasDataFrameItem.

This module defines the PandasDataFrameItem class,
which represents a pandas DataFrame item.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas

from skore.item.item import Item


class PandasDataFrameItem(Item):
    """
    A class to represent a pandas DataFrame item.

    This class encapsulates a pandas DataFrame along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        dataframe_dict: dict,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a PandasDataFrameItem.

        Parameters
        ----------
        dataframe_dict : dict
            The dict representation of the dataframe.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.dataframe_dict = dataframe_dict

    @cached_property
    def dataframe(self) -> pandas.DataFrame:
        """The pandas DataFrame."""
        import pandas

        return pandas.DataFrame.from_dict(self.dataframe_dict, orient="tight")

    @classmethod
    def factory(cls, dataframe: pandas.DataFrame) -> PandasDataFrameItem:
        """
        Create a new PandasDataFrameItem instance from a pandas DataFrame.

        Parameters
        ----------
        dataframe : pd.DataFrame
            The pandas DataFrame to store.

        Returns
        -------
        PandasDataFrameItem
            A new PandasDataFrameItem instance.
        """
        import pandas

        if not isinstance(dataframe, pandas.DataFrame):
            raise TypeError(f"Type '{dataframe.__class__}' is not supported.")

        instance = cls(dataframe_dict=dataframe.to_dict(orient="tight"))

        # add dataframe as cached property
        instance.dataframe = dataframe

        return instance



"""ItemRepository for managing storage and retrieval of items.

This module provides the ItemRepository class, which is responsible for
storing, retrieving, and deleting items in a storage system.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skore.item.item import Item
    from skore.persistence.abstract_storage import AbstractStorage


from skore.item.media_item import MediaItem
from skore.item.numpy_array_item import NumpyArrayItem
from skore.item.pandas_dataframe_item import PandasDataFrameItem
from skore.item.pandas_series_item import PandasSeriesItem
from skore.item.primitive_item import PrimitiveItem
from skore.item.sklearn_base_estimator_item import SklearnBaseEstimatorItem


class ItemRepository:
    """
    A repository for managing storage and retrieval of items.

    This class provides methods to get, put, and delete items from a storage system.
    """

    ITEM_CLASS_NAME_TO_ITEM_CLASS = {
        "MediaItem": MediaItem,
        "NumpyArrayItem": NumpyArrayItem,
        "PandasDataFrameItem": PandasDataFrameItem,
        "PandasSeriesItem": PandasSeriesItem,
        "PrimitiveItem": PrimitiveItem,
        "SklearnBaseEstimatorItem": SklearnBaseEstimatorItem,
    }

    def __init__(self, storage: AbstractStorage):
        """
        Initialize the ItemRepository with a storage system.

        Parameters
        ----------
        storage : AbstractStorage
            The storage system to be used by the repository.
        """
        self.storage = storage

    def get_item(self, key) -> Item:
        """
        Retrieve an item from storage.

        Parameters
        ----------
        key : Any
            The key used to identify the item in storage.

        Returns
        -------
        Item
            The retrieved item.
        """
        value = self.storage[key]
        item_class_name = value["item_class_name"]
        item_class = ItemRepository.ITEM_CLASS_NAME_TO_ITEM_CLASS[item_class_name]
        item = value["item"]

        return item_class(**item)

    def put_item(self, key, item: Item) -> None:
        """
        Store an item in storage.

        Parameters
        ----------
        key : Any
            The key to use for storing the item.
        item : Item
            The item to be stored.
        """
        item_parameters = item.__parameters__

        if key in self.storage:
            item_parameters["created_at"] = self.storage[key]["item"]["created_at"]

        self.storage[key] = {
            "item_class_name": item.__class__.__name__,
            "item": item_parameters,
        }

    def delete_item(self, key):
        """
        Delete an item from storage.

        Parameters
        ----------
        key : Any
            The key of the item to be deleted.
        """
        del self.storage[key]

    def keys(self) -> list[str]:
        """
        Get all keys of items stored in the repository.

        Returns
        -------
        list[str]
            A list of all keys in the storage.
        """
        return list(self.storage.keys())



"""The definition of API routes to list project items and get them."""

import base64
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Annotated, Any

from fastapi import APIRouter, HTTPException, Request, status
from fastapi.params import Depends
from fastapi.templating import Jinja2Templates

from skore.item.media_item import MediaItem
from skore.item.numpy_array_item import NumpyArrayItem
from skore.item.pandas_dataframe_item import PandasDataFrameItem
from skore.item.pandas_series_item import PandasSeriesItem
from skore.item.primitive_item import PrimitiveItem
from skore.item.sklearn_base_estimator_item import SklearnBaseEstimatorItem
from skore.project import Project
from skore.view.view import Layout, View

from .dependencies import get_static_path, get_templates

router = APIRouter(prefix="/project")


@dataclass
class SerializedItem:
    """Serialized item."""

    media_type: str
    value: Any
    updated_at: str
    created_at: str


@dataclass
class SerializedProject:
    """Serialized project, to be sent to the skore-ui."""

    items: dict[str, SerializedItem]
    views: dict[str, Layout]


def __serialize_project(project: Project) -> SerializedProject:
    views = {}
    for key in project.list_view_keys():
        views[key] = project.get_view(key).layout

    items = {}
    for key in project.list_item_keys():
        item = project.get_item(key)

        media_type = None
        if isinstance(item, PrimitiveItem):
            value = item.primitive
            media_type = "text/markdown"
        elif isinstance(item, NumpyArrayItem):
            value = item.array_list
            media_type = "text/markdown"
        elif isinstance(item, PandasDataFrameItem):
            value = item.dataframe_dict
            media_type = "application/vnd.dataframe+json"
        elif isinstance(item, PandasSeriesItem):
            value = item.series_list
            media_type = "text/markdown"
        elif isinstance(item, SklearnBaseEstimatorItem):
            value = item.estimator_html_repr
            media_type = "application/vnd.sklearn.estimator+html"
        elif isinstance(item, MediaItem):
            value = base64.b64encode(item.media_bytes).decode()
            media_type = item.media_type
        else:
            raise ValueError(f"Item {item} is not a known item type.")

        items[key] = SerializedItem(
            media_type=media_type,
            value=value,
            updated_at=item.updated_at,
            created_at=item.created_at,
        )

    return SerializedProject(
        items=items,
        views=views,
    )


@router.get("/items")
async def get_items(request: Request):
    """Serialize a project and send it."""
    project = request.app.state.project
    return __serialize_project(project)


@router.post("/views/share")
async def share_store(
    request: Request,
    key: str,
    templates: Annotated[Jinja2Templates, Depends(get_templates)],
    static_path: Annotated[Path, Depends(get_static_path)],
):
    """Serve an inlined shareable HTML page."""
    project = request.app.state.project

    try:
        project.get_view(key)
    except KeyError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="View not found"
        ) from None

    # Get static assets to inject them into the view template
    def read_asset_content(filename: str):
        with open(static_path / filename) as f:
            return f.read()

    script_content = read_asset_content("skore.umd.cjs")
    styles_content = read_asset_content("style.css")

    # Fill the Jinja context
    context = {
        "project": asdict(__serialize_project(project)),
        "selected_view": key,
        "script": script_content,
        "styles": styles_content,
    }

    # Render the template and send the result
    return templates.TemplateResponse(
        request=request, name="share.html.jinja", context=context
    )


@router.put("/views", status_code=status.HTTP_201_CREATED)
async def put_view(request: Request, key: str, layout: Layout):
    """Set the layout of the view corresponding to `key`.

    If the view corresponding to `key` does not exist, it will be created.
    """
    project: Project = request.app.state.project

    view = View(layout=layout)
    project.put_view(key, view)

    return __serialize_project(project)


@router.delete("/views", status_code=status.HTTP_202_ACCEPTED)
async def delete_view(request: Request, key: str):
    """Delete the view corresponding to `key`."""
    project: Project = request.app.state.project

    try:
        project.delete_view(key)
    except KeyError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="View not found"
        ) from None

    return __serialize_project(project)



"""Define a Project."""

import logging
from functools import singledispatchmethod
from pathlib import Path
from typing import Any, Literal, Union

from skore.item import (
    Item,
    ItemRepository,
    MediaItem,
    NumpyArrayItem,
    PandasDataFrameItem,
    PandasSeriesItem,
    PrimitiveItem,
    SklearnBaseEstimatorItem,
    object_to_item,
)
from skore.persistence.disk_cache_storage import DirectoryDoesNotExist, DiskCacheStorage
from skore.view.view import View
from skore.view.view_repository import ViewRepository

logger = logging.getLogger(__name__)


class ProjectPutError(Exception):
    """One more key-value pairs could not be saved in the Project."""


class Project:
    """A project is a collection of items that are stored in a storage."""

    def __init__(
        self,
        item_repository: ItemRepository,
        view_repository: ViewRepository,
    ):
        self.item_repository = item_repository
        self.view_repository = view_repository

    @singledispatchmethod
    def put(self, key: str, value: Any, on_error: Literal["warn", "raise"] = "warn"):
        """Add a value to the Project.

        If `on_error` is "raise", any error stops the execution. If `on_error`
        is "warn" (or anything other than "raise"), a warning is shown instead.

        Parameters
        ----------
        key : str
            The key to associate with `value` in the Project. Must be a string.
        value : Any
            The value to associate with `key` in the Project.
        on_error : "warn" or "raise", optional
            Upon error (e.g. if the key is not a string), whether to raise an error or
            to print a warning. Default is "warn".

        Raises
        ------
        ProjectPutError
            If the key-value pair cannot be saved properly, and `on_error` is "raise".
        """
        try:
            item = object_to_item(value)
            self.put_item(key, item)
        except (NotImplementedError, TypeError) as e:
            if on_error == "raise":
                raise ProjectPutError(
                    "Key-value pair could not be inserted in the Project"
                ) from e

            logger.warning(
                "Key-value pair could not be inserted in the Project "
                f"due to the following error: {e}"
            )

    @put.register
    def put_several(
        self, key_to_value: dict, on_error: Literal["warn", "raise"] = "warn"
    ):
        """Add several values to the Project.

        If `on_error` is "raise", the first error stops the execution (so the
        later key-value pairs will not be inserted). If `on_error` is "warn" (or
        anything other than "raise"), errors do not stop the execution, and are
        shown as they come as warnings; all the valid key-value pairs are inserted.

        Parameters
        ----------
        key_to_value : dict[str, Any]
            The key-value pairs to put in the Project. Keys must be strings.
        on_error : "warn" or "raise", optional
            Upon error (e.g. if a key is not a string), whether to raise an error or
            to print a warning. Default is "warn".

        Raises
        ------
        ProjectPutError
            If a key-value pair in `key_to_value` cannot be saved properly,
            and `on_error` is "raise".
        """
        for key, value in key_to_value.items():
            self.put(key, value, on_error=on_error)

    def put_item(self, key: str, item: Item):
        """Add an Item to the Project."""
        if not isinstance(key, str):
            raise TypeError(
                f"Key must be a string; key '{key}' is of type '{type(key)}'"
            )

        self.item_repository.put_item(key, item)

    def get(self, key: str) -> Any:
        """Get the value corresponding to `key` from the Project."""
        item = self.get_item(key)

        if isinstance(item, PrimitiveItem):
            return item.primitive
        elif isinstance(item, NumpyArrayItem):
            return item.array
        elif isinstance(item, PandasDataFrameItem):
            return item.dataframe
        elif isinstance(item, PandasSeriesItem):
            return item.series
        elif isinstance(item, SklearnBaseEstimatorItem):
            return item.estimator
        elif isinstance(item, MediaItem):
            return item.media_bytes
        else:
            raise ValueError(f"Item {item} is not a known item type.")

    def get_item(self, key: str) -> Item:
        """Add the Item corresponding to `key` from the Project."""
        return self.item_repository.get_item(key)

    def list_item_keys(self) -> list[str]:
        """List all item keys in the Project."""
        return self.item_repository.keys()

    def delete_item(self, key: str):
        """Delete an item from the Project."""
        self.item_repository.delete_item(key)

    def put_view(self, key: str, view: View):
        """Add a view to the Project."""
        self.view_repository.put_view(key, view)

    def get_view(self, key: str) -> View:
        """Get the view corresponding to `key` from the Project."""
        return self.view_repository.get_view(key)

    def delete_view(self, key: str):
        """Delete the view corresponding to `key` from the Project."""
        return self.view_repository.delete_view(key)

    def list_view_keys(self) -> list[str]:
        """List all view keys in the Project."""
        return self.view_repository.keys()


class ProjectLoadError(Exception):
    """Failed to load project."""


def load(project_name: Union[str, Path]) -> Project:
    """Load an existing Project given a project name or path."""
    # Transform a project name to a directory path:
    # - Resolve relative path to current working directory,
    # - Check that the file ends with the ".skore" extension,
    #    - If not provided, it will be automatically appended,
    # - If project name is an absolute path, we keep that path.

    path = Path(project_name).resolve()

    if path.suffix != ".skore":
        path = path.parent / (path.name + ".skore")

    if not Path(path).exists():
        raise ProjectLoadError(f"Project '{path}' does not exist: did you create it?")

    try:
        # FIXME should those hardcoded string be factorized somewhere ?
        item_storage = DiskCacheStorage(directory=Path(path) / "items")
        item_repository = ItemRepository(storage=item_storage)
        view_storage = DiskCacheStorage(directory=Path(path) / "views")
        view_repository = ViewRepository(storage=view_storage)
        project = Project(
            item_repository=item_repository,
            view_repository=view_repository,
        )
    except DirectoryDoesNotExist as e:
        missing_directory = e.args[0].split()[1]
        raise ProjectLoadError(
            f"Project '{path}' is corrupted: "
            f"directory '{missing_directory}' should exist. "
            "Consider re-creating the project."
        ) from e

    return project
